package com.flyfish.framework.compiler.core;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.BooleanUtils;
import org.apache.commons.lang3.StringUtils;

import javax.tools.JavaCompiler;
import javax.tools.JavaFileObject;
import javax.tools.StandardJavaFileManager;
import javax.tools.ToolProvider;
import java.io.IOException;
import java.util.Collections;

/**
 * In-memory compile Java source code as String.
 *
 * @author michael
 */
@Slf4j
class JavaStringCompiler implements VendorJavaCompiler {

    private final JavaCompiler compiler;
    private final MemoryJavaFileManager manager;

    JavaStringCompiler(DynamicClassLoader classLoader) {
        this.compiler = ToolProvider.getSystemJavaCompiler();
        StandardJavaFileManager stdManager = compiler.getStandardFileManager(null, null, null);
        this.manager = new MemoryJavaFileManager(stdManager, classLoader);
    }

    /**
     * Compile a Java source file in memory.
     *
     * @param fileName Java file name, e.g. "Test.java"
     * @param source   The source code as String.
     * @return The compiled results as Map that contains class name as key,
     * class binary as value.
     */
    @Override
    public String compile(String fileName, String source) {
        JavaFileObject javaFileObject = manager.makeStringSource(fileName, source);
        JavaCompiler.CompilationTask task = compiler.getTask(null, manager, null, null, null, Collections.singletonList(javaFileObject));
        String className = StringUtils.substringBeforeLast(fileName, ".");
        if (BooleanUtils.isTrue(task.call())) {
            return manager.getClassBytes().keySet().stream().filter(key -> key.endsWith(className))
                    .findFirst().orElse(null);
        }
        return null;
    }

    /**
     * 获取动态classLoader
     *
     * @return 结果
     */
    @Override
    public DynamicClassLoader getClassLoader() {
        return ClassLoaders.memory();
    }

    @Override
    public void close() {
        try {
            if (null != manager) {
                manager.close();
            }
            getClassLoader().close();
        } catch (IOException e) {
            log.error("关闭类编译器失败！", e);
        }
    }
}
