package com.flyfish.framework.query;

import com.flyfish.framework.query.Queries.Combinator;

import java.util.List;
import java.util.function.Supplier;

/**
 * 查询链
 * 此处为高度抽象，便于上层调用
 * 下层重写方法实现
 *
 * @param <C> 本类型泛型
 * @param <P> 参数泛型
 * @author wangyu
 */
public interface QueryChain<C extends QueryChain<C, P>, P> extends QueryDefinition {

    /**
     * 以且连接下一个字段
     *
     * @param column 列
     * @return 查询条件
     */
    QueryCondition<C> and(P column);

    /**
     * 直接拼接提供者，此处懒加载，最终build才会执行
     *
     * @param supplier 提供者
     * @param <V>      泛型，支持其他类型的链
     * @return 结果
     */
    <V extends QueryChain<V, ?>> C and(Supplier<QueryChain<V, ?>> supplier);

    /**
     * 条件列表
     *
     * @param chain 多个条件们
     * @return 结果
     */
    <V extends QueryChain<V, ?>> C and(QueryChain<V, ?> chain);

    /**
     * 多个嵌套子条件列表
     *
     * @param chains 多条链
     * @return 结果
     */
    default <V extends QueryChain<V, ?>> C and(List<QueryChain<V, ?>> chains) {
        return and(chains, Combinator.AND);
    }

    /**
     * 多个嵌套子条件列表
     *
     * @param chains     多条链
     * @param combinator 各个链条之间的连接方式
     * @return 结果
     */
    <V extends QueryChain<V, ?>> C and(List<QueryChain<V, ?>> chains, Combinator combinator);

    /**
     * 以或连接下一个字段
     *
     * @param column 列
     * @return 查询条件
     */
    QueryCondition<C> or(P column);

    /**
     * 以或连接条件列表
     *
     * @param chain 多个条件们
     * @return 结果
     */
    <V extends QueryChain<V, ?>> C or(QueryChain<V, ?> chain);

    /**
     * 以或连接嵌套多个子条件列表
     *
     * @param chains 多条链
     * @return 结果
     */
    default <V extends QueryChain<V, ?>> C or(List<QueryChain<V, ?>> chains) {
        return and(chains, Combinator.AND);
    }

    /**
     * 以或连接嵌套多个子条件列表
     *
     * @param chains     多条链
     * @param combinator 各个链条之间的连接方式
     * @return 结果
     */
    <V extends QueryChain<V, ?>> C or(List<QueryChain<V, ?>> chains, Combinator combinator);

    /**
     * 直接拼接提供者，此处懒加载，最终build才会执行
     *
     * @param supplier 提供者
     * @param <V>      泛型，支持其他类型的链
     * @return 结果
     */
    <V extends QueryChain<V, ?>> C or(Supplier<QueryChain<V, ?>> supplier);
}
