package com.flyfish.framework.query.chain;

import com.flyfish.framework.query.Queries;
import com.flyfish.framework.query.QueryChain;
import com.flyfish.framework.query.QueryCondition;
import com.flyfish.framework.query.QueryMutation;
import com.flyfish.framework.query.holder.QueryChainHolder;
import com.flyfish.framework.query.spi.adaptor.CriteriaAdaptor;
import lombok.Getter;
import org.springframework.data.util.CastUtils;

import java.util.List;
import java.util.function.Supplier;

/**
 * 默认的查询定义，交给holder提供构建
 * 在执行时，我们并不知道adaptor是谁，只有最后构建阶段才能明确
 *
 * @author wangyu
 */
@Getter
abstract class DefaultQueryChain<C extends QueryChain<C, P>, P> implements QueryChain<C, P> {

    protected final DefaultQueryChainHolder<?> holder = new DefaultQueryChainHolder<>();

    /**
     * 以且的关系进行拼接
     *
     * @return 结果
     */
    protected QueryChainHolder<?> and() {
        return holder.link(adaptor -> adaptor::and);
    }

    /**
     * 以或的关系进行拼接
     *
     * @return 结果
     */
    protected QueryChainHolder<?> or() {
        return holder.link(adaptor -> adaptor::or);
    }

    /**
     * 创建条件，使用默认实现的条件
     *
     * @param column 列名
     * @return 结果
     */
    protected QueryCondition<C> createCondition(String column) {
        return new DefaultQueryCondition<>(self(), holder, column);
    }

    /**
     * 返回自身真实类型
     *
     * @return 结果
     */
    private C self() {
        return CastUtils.cast(this);
    }

    /**
     * 构建结果
     *
     * @param adaptor 查询适配器
     * @return 构建结果
     */
    @Override
    @SuppressWarnings("unchecked")
    public <T> T build(CriteriaAdaptor<T> adaptor) {
        QueryChainHolder<T> casted = (QueryChainHolder<T>) holder;
        return casted.get(adaptor);
    }

    /**
     * 修改，此修改会直接接着查询条件进行拼接
     *
     * @return 修改句柄
     */
    @Override
    public QueryMutation mutate() {
        return new DefaultQueryMutation(this, false);
    }

    /**
     * 会直接添加嵌套查询到当前查询末尾
     *
     * @return 修改句柄
     */
    @Override
    public QueryMutation within() {
        return new DefaultQueryMutation(this, true);
    }

    /**
     * 判断查询条件是否为空
     *
     * @return 结果
     */
    @Override
    public boolean isEmpty() {
        return holder.isEmpty();
    }

    /**
     * 直接拼接提供者，此处懒加载，最终build才会执行
     *
     * @param supplier 提供者
     * @return 结果
     */
    @Override
    public <V extends QueryChain<V, ?>> C and(Supplier<QueryChain<V, ?>> supplier) {
        and().with(adaptor -> supplier.get().build(adaptor));
        return self();
    }

    /**
     * 条件列表
     *
     * @param chain 多个条件们
     * @return 结果
     */
    @Override
    public <V extends QueryChain<V, ?>> C and(QueryChain<V, ?> chain) {
        and().with(chain::build);
        return self();
    }

    /**
     * 多个嵌套子条件列表
     *
     * @param chains     多条链
     * @param combinator 各个链条之间的连接方式
     * @return 结果
     */
    @Override
    public <V extends QueryChain<V, ?>> C and(List<QueryChain<V, ?>> chains, Queries.Combinator combinator) {
        return and(combine(chains, combinator));
    }

    /**
     * 以或连接条件列表
     *
     * @param chain 多个条件们
     * @return 结果
     */
    @Override
    public <V extends QueryChain<V, ?>> C or(QueryChain<V, ?> chain) {
        or().with(chain::build);
        return self();
    }

    /**
     * 以或连接嵌套多个子条件列表
     *
     * @param chains     多条链
     * @param combinator 各个链条之间的连接方式
     * @return 结果
     */
    @Override
    public <V extends QueryChain<V, ?>> C or(List<QueryChain<V, ?>> chains, Queries.Combinator combinator) {
        return or(combine(chains, combinator));
    }

    /**
     * 直接拼接提供者，此处懒加载，最终build才会执行
     *
     * @param supplier 提供者
     * @return 结果
     */
    @Override
    public <V extends QueryChain<V, ?>> C or(Supplier<QueryChain<V, ?>> supplier) {
        or().with(adaptor -> supplier.get().build(adaptor));
        return self();
    }

    /**
     * 结合多个查询链，并根据组合方式进行处理
     *
     * @param chains     查询链集合
     * @param combinator 组合方式
     * @return 结果
     */
    private <V extends QueryChain<V, ?>> C combine(List<QueryChain<V, ?>> chains, Queries.Combinator combinator) {
        return chains.stream().reduce((left, right) -> combinator == Queries.Combinator.OR ? left.or(right) : left.and(right))
                .map(this::and)
                .orElse(self());
    }
}
