package com.flyfish.framework.query.chain;

import com.flyfish.framework.query.Queries;
import com.flyfish.framework.query.QueryChain;
import com.flyfish.framework.query.QueryCondition;
import com.flyfish.framework.query.holder.QueryChainHolder;
import lombok.RequiredArgsConstructor;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;

import java.util.Collection;
import java.util.List;

/**
 * 条件操作实现
 *
 * @author wangyu
 */
@RequiredArgsConstructor
class DefaultQueryCondition<C extends QueryChain<C, ?>> implements QueryCondition<C> {

    private final C chain;
    private final QueryChainHolder<?> holder;
    private final String column;

    /**
     * 相等判定
     *
     * @param value 值
     * @return 结果
     */
    @Override
    public C eq(Object value) {
        if (null != value) {
            holder.with(adaptor -> adaptor.eq(column, value));
        }
        return chain;
    }

    /**
     * 判定某列的值中存在指定值，特指json array数据类型且子类型中带有id的场景
     * 用于兼容mongodb查询，mysql查询使用JSON_CONTAINS进行判定。
     * 当且仅当mongodb会拼接.$id，mysql一律匹配id字段
     * <p>
     * 等价于 .eq(value)
     *
     * @param value 值
     * @return 结果
     */
    @Override
    public C hasId(Object value) {
        if (null != value) {
            holder.with(adaptor -> adaptor.hasId(column, value));
        }
        return chain;
    }

    /**
     * 对于mongodb，自动处理。对于关系型数据库，代表json array中是否包含对应值，值仅支持基本数据类型
     * <p>
     * 等价于 .eq(value)
     *
     * @param value 基本数据类型的值
     * @return 结果
     */
    @Override
    public C has(Object value) {
        if (null != value) {
            holder.with(adaptor -> adaptor.has(column, value));
        }
        return chain;
    }

    /**
     * 不等判定
     *
     * @param value 值
     * @return 结果
     */
    @Override
    public C ne(Object value) {
        if (null != value) {
            holder.with(adaptor -> adaptor.ne(column, value));
        }
        return chain;
    }

    /**
     * 值介于两者之间
     *
     * @param items 双值列表
     * @return 结果
     */
    @Override
    public C between(List<?> items) {
        if (CollectionUtils.size(items) == 2) {
            holder.with(adaptor -> adaptor.between(column, items));
        }
        return chain;
    }

    /**
     * 模糊匹配，这里是全模糊
     *
     * @param keyword 查询关键字
     * @return 结果
     */
    @Override
    public C like(String keyword) {
        if (StringUtils.isNotBlank(keyword)) {
            holder.with(adaptor -> adaptor.like(column, keyword, Queries.Direction.ALL));
        }
        return chain;
    }

    /**
     * 根据指定的方向进行模糊查询
     *
     * @param keyword   关键字
     * @param direction 方向，可以匹配开头和结尾
     * @return 结果
     */
    @Override
    public C like(String keyword, Queries.Direction direction) {
        if (StringUtils.isNotBlank(keyword)) {
            holder.with(adaptor -> adaptor.like(column, keyword, direction));
        }
        return chain;
    }

    /**
     * 判定为空
     *
     * @return 结果
     */
    @Override
    public C isNull() {
        holder.with(adaptor -> adaptor.isNull(column));
        return chain;
    }

    /**
     * 包含在内
     *
     * @param list 集合
     * @return 结果
     */
    @Override
    public C in(Collection<?> list) {
        if (CollectionUtils.isNotEmpty(list)) {
            holder.with(adaptor -> adaptor.in(column, list));
        }
        return chain;
    }

    /**
     * 包含在内
     * 注意，需要根据字段类型推断。
     * 如果是json数组，需要进行双向匹配
     *
     * @param values 值们
     * @return 结果
     */
    @Override
    public C in(Object... values) {
        if (ArrayUtils.isNotEmpty(values)) {
            holder.with(adaptor -> adaptor.in(column, values));
        }
        return chain;
    }

    /**
     * 不包含在内
     * 注意，需要根据字段类型推断。
     * 如果是json数组，需要进行双向匹配
     *
     * @param values 值们
     * @return 结果
     */
    @Override
    public C nin(Object... values) {
        if (ArrayUtils.isNotEmpty(values)) {
            holder.with(adaptor -> adaptor.nin(column, values));
        }
        return chain;
    }

    /**
     * 不包含在内
     *
     * @param values 集合
     * @return 结果
     */
    @Override
    public C nin(Collection<?> values) {
        if (CollectionUtils.isNotEmpty(values)) {
            holder.with(adaptor -> adaptor.nin(column, values));
        }
        return chain;
    }

    /**
     * 小于指定的值
     *
     * @param value 类型
     * @return 结果
     */
    @Override
    public C lt(Object value) {
        if (null != value) {
            holder.with(adaptor -> adaptor.lt(column, value));
        }
        return chain;
    }

    /**
     * 小于等于指定的值
     *
     * @param value 类型
     * @return 结果
     */
    @Override
    public C lte(Object value) {
        if (null != value) {
            holder.with(adaptor -> adaptor.lte(column, value));
        }
        return chain;
    }

    /**
     * 大于指定的值
     *
     * @param value 类型
     * @return 结果
     */
    @Override
    public C gt(Object value) {
        if (null != value) {
            holder.with(adaptor -> adaptor.gt(column, value));
        }
        return chain;
    }

    /**
     * 大于等于指定的值
     *
     * @param value 类型
     * @return 结果
     */
    @Override
    public C gte(Object value) {
        if (null != value) {
            holder.with(adaptor -> adaptor.gte(column, value));
        }
        return chain;
    }
}
