package com.flyfish.framework.r2dbc.config.callback;

import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcAssociation;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcCollection;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcRelation;
import com.flyfish.framework.r2dbc.metadata.visitor.MetadataHandler;
import com.flyfish.framework.repository.DefaultReactiveRepository;
import lombok.RequiredArgsConstructor;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.data.mapping.PersistentEntity;
import org.springframework.data.mapping.PersistentPropertyAccessor;
import org.springframework.data.r2dbc.core.R2dbcEntityOperations;
import org.springframework.data.r2dbc.core.StatementMapper;
import org.springframework.data.relational.core.query.Criteria;
import org.springframework.r2dbc.core.DatabaseClient;
import org.springframework.r2dbc.core.Parameter;
import org.springframework.r2dbc.core.PreparedOperation;
import reactor.core.publisher.Mono;

import java.util.List;
import java.util.function.Supplier;
import java.util.stream.Collectors;

@RequiredArgsConstructor
class AssociationWriter implements MetadataHandler {

    private final RepositoryProvider repositories;

    private final Supplier<R2dbcEntityOperations> operations;

    /**
     * 处理association
     *
     * @param association 一对一关联
     * @param accessor    对象访问器
     * @return 结果
     */
    @Override
    public Mono<Domain> handleAssociation(R2dbcAssociation association, PersistentPropertyAccessor<Domain> accessor) {
        Domain entity = accessor.getBean();
        // 获取目标值
        Domain value = association.getAccessor().get(entity);
        if (null == value) return null;
        // 准备仓库
        DefaultReactiveRepository<Domain> repository = repositories.get(association.getEntity());
        // 执行逻辑
        if (association.isInner()) {
            // 保存内部关联对象，并设置关联值。如果有id，更新，没有id，插入
            if (association.getEntity().isNew(value)) {
                // 先保存，再设置
                return repository.insert(value).map(saved -> {
                    // 设置内部id关联
                    accessor.setProperty(association.getField(), saved.getId());
                    return entity;
                });
            } else {
                // 设置内部id关联
                accessor.setProperty(association.getField(), value.getId());
                // 尝试保存
                return repository.save(value).thenReturn(entity);
            }
        } else {
            // 保存外部关联对象，并更新外部关联当前实体的id
            PersistentPropertyAccessor<Domain> otherAccessor = association.getEntity().getPropertyAccessor(value);
            // 先设置目标id
            otherAccessor.setProperty(association.getField(), entity.getId());
            // 再保存
            return repository.save(value).thenReturn(entity);
        }
    }

    /**
     * 处理collection
     *
     * @param collection 一对多关联
     * @param accessor   对象访问器
     * @return 结果
     */
    @Override
    public Mono<Domain> handleCollection(R2dbcCollection collection, PersistentPropertyAccessor<Domain> accessor) {
        Domain entity = accessor.getBean();
        // 获取目标值
        List<Domain> value = collection.getAccessor().get(entity);
        if (CollectionUtils.isEmpty(value)) return null;
        // 获取目标实体信息
        PersistentEntity<?, ?> targetEntity = collection.getEntity();
        // 保存外部关联对象，并更新外部关联当前实体的id
        value.forEach(item -> {
            PersistentPropertyAccessor<Domain> otherAccessor = collection.getEntity().getPropertyAccessor(item);
            otherAccessor.setProperty(collection.getField(), entity.getId());
        });
        // 准备仓库
        DefaultReactiveRepository<Domain> repository = repositories.get(targetEntity);
        // 批量保存
        return repository.saveAll(value).collectList()
                .map(list -> collection.getAccessor().set(entity, list));
    }

    /**
     * 处理relation
     *
     * @param relation 多对多关联
     * @param accessor 访问器
     * @return 结果
     */
    @Override
    public Mono<Domain> handleRelation(R2dbcRelation relation, PersistentPropertyAccessor<Domain> accessor) {
        // 保存多对多关系，需要先分别插入两边的数据，然后在关联表新增或更新
        Domain entity = accessor.getBean();
        // 获取目标值
        List<Domain> value = relation.getAccessor().get(entity);
        if (CollectionUtils.isEmpty(value)) return null;
        // 获取目标实体信息
        PersistentEntity<?, ?> targetEntity = relation.getEntity();
        // 准备仓库
        DefaultReactiveRepository<Domain> repository = repositories.get(targetEntity);
        // 批量保存
        return repository.saveAll(value).collectList()
                .flatMap(list -> {
                    // 回填数据
                    relation.getAccessor().set(entity, list);
                    // 关系表保存
                    return saveRelationIds(relation, entity);
                });
    }

    /**
     * 保存关系表
     *
     * @param relation 关系信息
     * @param entity   实体信息
     */
    @SuppressWarnings("deprecation")
    private Mono<Domain> saveRelationIds(R2dbcRelation relation, Domain entity) {
        // 准备查询器
        R2dbcEntityOperations entityOperations = operations.get();
        StatementMapper statementMapper = entityOperations.getDataAccessStrategy().getStatementMapper();
        DatabaseClient databaseClient = entityOperations.getDatabaseClient();
        // 构建保存语句，先删后增
        StatementMapper.DeleteSpec deleteSpec = statementMapper.createDelete(relation.getTableName())
                .withCriteria(Criteria.where(relation.getField()).is(entity.getId()));
        PreparedOperation<?> deletion = statementMapper.getMappedObject(deleteSpec);
        Mono<Domain> deleting = execute(databaseClient, deletion, entity);
        // 构建插入语句，多条
        List<Domain> items = relation.getAccessor().get(entity);
        List<Mono<Domain>> insertions = items.stream()
                .map(item -> {
                    StatementMapper.InsertSpec insertion = statementMapper.createInsert(relation.getTableName())
                            .withColumn(relation.getField(), Parameter.from(entity.getId()))
                            .withColumn(relation.getForeignField(), Parameter.from(item.getId()));
                    return statementMapper.getMappedObject(insertion);
                })
                .map(operation -> execute(databaseClient, operation, entity))
                .collect(Collectors.toList());
        // 执行查询
        return deleting.flatMap(list -> Mono.zip(insertions, results -> entity));
    }

    /**
     * 执行更新语句的逻辑，保证多行每一行都有执行到
     *
     * @param databaseClient 数据库客户端
     * @param operation      预编译的sql
     * @param entity         实体
     * @return 结果
     */
    private Mono<Domain> execute(DatabaseClient databaseClient, PreparedOperation<?> operation, Domain entity) {
        return databaseClient.sql(operation)
                .fetch()
                .rowsUpdated()
                .thenReturn(entity);
    }
}
