package com.flyfish.framework.logging.service;

import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.domain.po.User;
import com.flyfish.framework.logging.config.LoggingTextRegistry;
import com.flyfish.framework.logging.domain.Log;
import com.flyfish.framework.logging.domain.LogType;
import com.flyfish.framework.service.AuthenticationLogger;
import com.flyfish.framework.utils.AuthenticationMessages;
import com.flyfish.framework.utils.UserUtils;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.stereotype.Service;
import reactor.core.publisher.Mono;

import javax.annotation.Resource;
import java.time.LocalDateTime;
import java.util.Optional;

/**
 * 简单的认证日志
 *
 * @author wangyu
 */
@Service
public class SimpleAuthenticationLogger implements AuthenticationLogger {

    @Resource
    private LogService logService;
    @Resource
    private LoggingTextRegistry registry;

    /**
     * 记录成功
     *
     * @param user 用户
     */
    @Override
    public Mono<Void> success(UserDetails user) {
        return saveSuccess("LOGIN", "登录成功", user);
    }

    /**
     * 记录失败
     *
     * @param user      表单用户
     * @param exception 错误异常
     */
    @Override
    public Mono<Void> failure(User user, AuthenticationException exception) {
        return saveError("LOGIN", "登录失败", user, AuthenticationMessages.message(exception));
    }

    /**
     * 记录退出
     *
     * @param user 用户
     */
    @Override
    public Mono<Void> logout(UserDetails user) {
        return saveSuccess("LOGOUT", "退出成功", user);
    }

    /**
     * 操作成功记录
     *
     * @param business 业务
     * @param message  信息
     * @param user     用户
     */
    private Mono<Void> saveSuccess(String business, String message, UserDetails user) {
        User raw = UserUtils.toUser(user);
        Log log = new Log();
        log.setType(LogType.AUTHENTICATION);
        log.setSignature(business);
        log.setSuccess(true);
        log.setBody("为了安全，本内容隐藏");
        log.setModule("登录模块");
        log.setBusiness(registry.text(business));
        log.setResponse(message);
        log.setOperator(Optional.ofNullable(raw).map(Domain::getName).orElse("未知"));
        log.setPeriod(0L);
        log.setStartTime(LocalDateTime.now());
        // 写入日志
        return logService.create(log).then();
    }

    /**
     * 保存失败日志
     *
     * @param business 业务
     * @param message  消息
     * @param error    错误
     */
    private Mono<Void> saveError(String business, String message, User user, String error) {
        Log log = new Log();
        log.setType(LogType.AUTHENTICATION);
        log.setSignature(business);
        log.setSuccess(false);
        log.setBody("某人尝试使用用户名'" + user.getUsername() + "'，密码：******，进行登录");
        log.setModule("登录模块");
        log.setBusiness(registry.text(business));
        log.setError(error);
        log.setResponse(message);
        log.setOperator(user.getName());
        log.setPeriod(0L);
        log.setStartTime(LocalDateTime.now());
        // 写入日志
        return logService.create(log).then();
    }

}
