package com.flyfish.framework.config;

import com.flyfish.framework.beans.annotations.EnableRestBeanDetect;
import com.flyfish.framework.beans.enums.RestBeanCandidate;
import com.flyfish.framework.beans.meta.RestBean;
import com.flyfish.framework.beans.repository.RepositoryRegistrar;
import com.flyfish.framework.beans.resolver.DynamicRestBeanResolver;
import com.flyfish.framework.compiler.DynamicJavaCompiler;
import com.flyfish.framework.compiler.support.JavaSource;
import com.flyfish.framework.compiler.template.TemplateCompiler;
import com.flyfish.framework.domain.base.Vo;
import com.flyfish.framework.repository.DefaultReactiveRepository;
import com.flyfish.framework.repository.DefaultRepository;
import com.flyfish.framework.utils.Assert;
import com.flyfish.framework.utils.ClassPathResourceScanner;
import com.flyfish.framework.utils.StringFormats;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.support.BeanDefinitionBuilder;
import org.springframework.beans.factory.support.BeanDefinitionRegistry;
import org.springframework.beans.factory.support.BeanNameGenerator;
import org.springframework.context.EnvironmentAware;
import org.springframework.context.ResourceLoaderAware;
import org.springframework.context.annotation.ImportBeanDefinitionRegistrar;
import org.springframework.core.env.Environment;
import org.springframework.core.io.ResourceLoader;
import org.springframework.core.type.AnnotationMetadata;

import java.io.IOException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static com.flyfish.framework.beans.enums.RestBeanCandidate.*;

/**
 * rest bean的自动配置
 *
 * @author wangyu
 */
@Slf4j
public class RestBeanAutoConfigure implements ImportBeanDefinitionRegistrar, ResourceLoaderAware, EnvironmentAware {

    // 预编译模板
    private final Map<RestBeanCandidate, Function<JavaSource, String>> templates = Arrays.stream(values())
            .collect(Collectors.toMap(key -> key, key -> TemplateCompiler.compile("/templates/" + key.getName() + ".tpl")));
    // 存放父类值的映射
    private final Map<RestBeanCandidate, Function<RestBean, Class<?>>> superClasses;
    // 默认的父类行为
    private final Function<RestBean, Class<?>> noOp = a -> Object.class;
    // 编译器引用
    private final DynamicJavaCompiler compiler = DynamicJavaCompiler.delegate();

    // 资源加载器
    private ResourceLoader resourceLoader;
    // 环境变量
    private Environment environment;

    public RestBeanAutoConfigure() {
        superClasses = new HashMap<>();
        superClasses.put(CONTROLLER, RestBean::controllerClass);
        superClasses.put(REPOSITORY, RestBean::repoClass);
        superClasses.put(SERVICE, RestBean::serviceClass);
    }

    /**
     * 注册bean的支持，能够处理注解信息，注册bean等
     *
     * @param metadata 导入类的元数据
     * @param registry 注册器
     */
    @Override
    public void registerBeanDefinitions(AnnotationMetadata metadata, BeanDefinitionRegistry registry, BeanNameGenerator beanNameGenerator) {
        log.info("开始注册rest bean...");
        // 获取包路径元数据
        Set<String> basePackages = getBasePackages(metadata);
        // repo注册器
        RepositoryRegistrar repositoryRegistrar = RepositoryRegistrar.newInstance(metadata)
                .init(metadata, resourceLoader, environment, registry, beanNameGenerator);
        // 寻找基本包路径
        String basePackage = basePackages.stream().findFirst().orElse("com.flyfish.project");
        // 获取被注解的类，开始编译
        Set<Class<?>> classes = ClassPathResourceScanner.forAnnotation(RestBean.class).scan(basePackages);
        // 并发编译，快速接入，准备注册的repo
        Map<RestBeanCandidate, List<String>> compiledClasses = compile(basePackage, classes);
        // 从repo开始，注册bean
        compiledClasses.getOrDefault(REPOSITORY, Collections.emptyList())
                .stream()
                .map(compiler::load)
                .forEach(repositoryRegistrar::register);
        compiledClasses.getOrDefault(SERVICE, Collections.emptyList())
                .forEach(clazz -> registerBean(clazz, registry));
        compiledClasses.getOrDefault(CONTROLLER, Collections.emptyList())
                .forEach(clazz -> registerBean(clazz, registry));
        compiledClasses.getOrDefault(VIEW_CONTROLLER, Collections.emptyList())
                .forEach(clazz -> registerBean(clazz, registry));
        compiledClasses.getOrDefault(REACTIVE_VIEW_CONTROLLER, Collections.emptyList())
                .forEach(clazz -> registerBean(clazz, registry));
        log.info("结束注册rest bean...");
    }

    /**
     * 解析并编译带注解的rest bean
     *
     * @param basePackage 基本路径
     * @param classes     类集合
     * @return 结果
     */
    private Map<RestBeanCandidate, List<String>> compile(String basePackage, Set<Class<?>> classes) {
        Map<RestBeanCandidate, List<String>> compiled = new ConcurrentHashMap<>();
        classes.forEach(clazz -> {
            RestBean restBean = clazz.getDeclaredAnnotation(RestBean.class);
            if (null != restBean) {
                // 注册bean信息
                DynamicRestBeanResolver.register(restBean.value(), clazz);
                // 判断是否有vo
                boolean hasVo = !Vo.class.equals(restBean.listViewClass());
                // 创建java source信息
                JavaSource source = JavaSource.create()
                        .setBeanClass(clazz.getCanonicalName())
                        .setQueryBeanClass(restBean.queryClass().getCanonicalName())
                        .setListViewClass(hasVo ? restBean.listViewClass().getCanonicalName() : null)
                        .setPackageName(basePackage)
                        .setUri(makePath(clazz.getSimpleName(), restBean.value()));
                // 分别生成实现类，从repo到controller
                templates.forEach((type, template) -> {
                    // 当且仅当存在vo时，才编译view-controller
                    if (testNotSupport(restBean, type, hasVo)) {
                        return;
                    }
                    source.setSuperClass(superClasses.getOrDefault(type, noOp).apply(restBean).getCanonicalName());
                    source.setClassName(clazz.getSimpleName() + type.getName());
                    try {
                        log.info("尝试编译{}", source.getClassName());
                        String className = compiler.compile(source.getClassName() + ".java", template.apply(source));
                        Assert.hasText(className, "【java编译器】源码中没有可以编译的类！");
                        compiled.computeIfAbsent(type, k -> new ArrayList<>()).add(className);
                    } catch (IOException e) {
                        log.error("编译{}时发生异常！", source.getClassName(), e);
                    } catch (Exception e) {
                        log.error("编译{}警告！{}", source.getClassName(), e.getMessage(), e);
                    }
                });
            }
        });
        return compiled;
    }

    /**
     * 注册bean，如果是controller，自动注册controller
     *
     * @param className class，要注册的class
     * @param registry  注册器，注册bean用
     */
    private void registerBean(String className, BeanDefinitionRegistry registry) {
        Class<?> clazz = compiler.load(className);
        Assert.notNull(clazz, "类加载失败:" + className);
        //生成BeanDefinition并注册到容器中
        BeanDefinitionBuilder builder = BeanDefinitionBuilder.genericBeanDefinition(clazz);
        BeanDefinition beanDefinition = builder.getRawBeanDefinition();
        //设置当前bean定义对象是单利的
        beanDefinition.setScope("singleton");
        //将变量首字母置小写
        String beanName = StringUtils.uncapitalize(clazz.getSimpleName());
        log.info("注册bean：{}", beanName);
        registry.registerBeanDefinition(beanName, builder.getBeanDefinition());
    }

    /**
     * 创建路径，缺省加s，对于错误拼写需要直接指定
     *
     * @param className 类名
     * @param uri       路径
     * @return 结果
     */
    private String makePath(String className, String uri) {
        return StringUtils.isBlank(uri) ? (StringFormats.camel2Line(className) + "s") : uri;
    }

    /**
     * 获取基本扫描包路径
     *
     * @param metadata 元数据
     * @return 结果
     */
    private Set<String> getBasePackages(AnnotationMetadata metadata) {
        // 获取启动类的注解配置
        Map<String, Object> attrs = metadata.getAnnotationAttributes(EnableRestBeanDetect.class.getCanonicalName(), true);
        Assert.isTrue(null != attrs && attrs.containsKey("basePackages"), "【rest配置】未指定包路径！");
        String[] packages = (String[]) attrs.get("basePackages");
        if (ArrayUtils.isEmpty(packages)) {
            return Stream.of(metadata.getClassName()).collect(Collectors.toSet());
        } else {
            return Arrays.stream(packages).filter(StringUtils::isNotBlank).collect(Collectors.toSet());
        }
    }

    @Override
    public void setResourceLoader(ResourceLoader resourceLoader) {
        this.resourceLoader = resourceLoader;
    }

    @Override
    public void setEnvironment(Environment environment) {
        this.environment = environment;
    }

    /**
     * 测试是否支持vo
     *
     * @param bean  定义
     * @param type  注入类型
     * @param hasVo 是否有vo
     * @return 结果
     */
    private boolean testNotSupport(RestBean bean, RestBeanCandidate type, boolean hasVo) {
        Class<?> superClass = superClasses.get(REPOSITORY).apply(bean);
        switch (type) {
            case VIEW_CONTROLLER:
                return !hasVo || !DefaultRepository.class.isAssignableFrom(superClass);
            case REACTIVE_VIEW_CONTROLLER:
                return !hasVo || !DefaultReactiveRepository.class.isAssignableFrom(superClass);
            default:
                return false;
        }
    }
}
