package com.flyfish.framework.configuration.resolver;

import com.flyfish.framework.configuration.annotations.ValidRequestBody;
import com.flyfish.framework.validation.groups.CodeValid;
import com.flyfish.framework.validation.groups.NameValid;
import com.flyfish.framework.validation.spi.ConditionalGroup;
import com.flyfish.framework.validation.spi.ConditionalValidGroupProvider;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.springframework.core.*;
import org.springframework.core.annotation.AnnotationUtils;
import org.springframework.core.codec.DecodingException;
import org.springframework.core.codec.Hints;
import org.springframework.core.io.buffer.DataBuffer;
import org.springframework.core.io.buffer.DataBufferUtils;
import org.springframework.data.util.CastUtils;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.http.codec.HttpMessageReader;
import org.springframework.http.server.reactive.ServerHttpRequest;
import org.springframework.http.server.reactive.ServerHttpResponse;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;
import org.springframework.util.Assert;
import org.springframework.web.bind.support.WebExchangeBindException;
import org.springframework.web.bind.support.WebExchangeDataBinder;
import org.springframework.web.reactive.BindingContext;
import org.springframework.web.reactive.result.method.annotation.AbstractMessageReaderArgumentResolver;
import org.springframework.web.server.ServerWebExchange;
import org.springframework.web.server.ServerWebInputException;
import org.springframework.web.server.UnsupportedMediaTypeStatusException;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.*;
import java.util.stream.Collectors;

/**
 * 私有校验逻辑支持的body
 *
 * @author wangyu
 * 支持条件校验，更高级的骚操作，大量重写
 */
public class ValidRequestBodyMethodArgumentResolver extends AbstractMessageReaderArgumentResolver {

    private static final Set<HttpMethod> SUPPORTED_METHODS =
            EnumSet.of(HttpMethod.POST, HttpMethod.PUT, HttpMethod.PATCH);

    private final List<MediaType> supportedMediaTypes;

    public ValidRequestBodyMethodArgumentResolver(List<HttpMessageReader<?>> readers, ReactiveAdapterRegistry registry) {
        super(readers, registry);
        supportedMediaTypes = readers.stream()
                .flatMap(converter -> converter.getReadableMediaTypes().stream())
                .collect(Collectors.toList());
    }

    @Override
    public boolean supportsParameter(MethodParameter parameter) {
        return parameter.hasParameterAnnotation(ValidRequestBody.class);
    }

    @Override
    public Mono<Object> resolveArgument(
            MethodParameter param, BindingContext bindingContext, ServerWebExchange exchange) {

        ValidRequestBody ann = param.getParameterAnnotation(ValidRequestBody.class);
        Assert.state(ann != null, "No RequestBody annotation");
        return readBody(param, ann.required(), bindingContext, exchange);
    }

    @Override
    protected Mono<Object> readBody(MethodParameter bodyParam, @Nullable MethodParameter actualParam,
                                    boolean isBodyRequired, BindingContext bindingContext, ServerWebExchange exchange) {

        ResolvableType bodyType = ResolvableType.forMethodParameter(bodyParam);
        ResolvableType actualType = (actualParam != null ? ResolvableType.forMethodParameter(actualParam) : bodyType);
        Class<?> resolvedType = bodyType.resolve();
        ReactiveAdapter adapter = (resolvedType != null ? getAdapterRegistry().getAdapter(resolvedType) : null);
        ResolvableType elementType = (adapter != null ? bodyType.getGeneric() : bodyType);
        isBodyRequired = isBodyRequired || (adapter != null && !adapter.supportsEmpty());

        ServerHttpRequest request = exchange.getRequest();
        ServerHttpResponse response = exchange.getResponse();

        MediaType contentType = request.getHeaders().getContentType();
        MediaType mediaType = (contentType != null ? contentType : MediaType.APPLICATION_OCTET_STREAM);

        if (mediaType.isCompatibleWith(MediaType.APPLICATION_FORM_URLENCODED)) {
            return Mono.error(new IllegalStateException(
                    "In a WebFlux application, form data is accessed via ServerWebExchange.getFormData()."));
        }

        if (logger.isDebugEnabled()) {
            logger.debug(exchange.getLogPrefix() + (contentType != null ?
                    "Content-Type:" + contentType :
                    "No Content-Type, using " + MediaType.APPLICATION_OCTET_STREAM));
        }

        for (HttpMessageReader<?> reader : getMessageReaders()) {
            if (reader.canRead(elementType, mediaType)) {
                Map<String, Object> readHints = Hints.from(Hints.LOG_PREFIX_HINT, exchange.getLogPrefix());
                if (adapter != null && adapter.isMultiValue()) {
                    if (logger.isDebugEnabled()) {
                        logger.debug(exchange.getLogPrefix() + "0..N [" + elementType + "]");
                    }
                    Flux<?> flux = reader.read(actualType, elementType, request, response, readHints);
                    flux = flux.onErrorResume(ex -> Flux.error(handleReadError(bodyParam, ex)));
                    if (isBodyRequired) {
                        flux = flux.switchIfEmpty(Flux.error(() -> handleMissingBody(bodyParam)));
                    }
                    flux = flux.doOnNext(target -> validate(target, bodyParam, bindingContext, exchange));
                    return Mono.just(adapter.fromPublisher(flux));
                } else {
                    // Single-value (with or without reactive type wrapper)
                    if (logger.isDebugEnabled()) {
                        logger.debug(exchange.getLogPrefix() + "0..1 [" + elementType + "]");
                    }
                    Mono<?> mono = reader.readMono(actualType, elementType, request, response, readHints);
                    mono = mono.onErrorResume(ex -> Mono.error(handleReadError(bodyParam, ex)));
                    if (isBodyRequired) {
                        mono = mono.switchIfEmpty(Mono.error(() -> handleMissingBody(bodyParam)));
                    }
                    mono = mono.doOnNext(target -> validate(target, bodyParam, bindingContext, exchange));
                    return (adapter != null ? Mono.just(adapter.fromPublisher(mono)) : Mono.from(mono));
                }
            }
        }

        // No compatible reader but body may be empty..

        HttpMethod method = request.getMethod();
        if (contentType == null && method != null && SUPPORTED_METHODS.contains(method)) {
            Flux<DataBuffer> body = request.getBody().doOnNext(buffer -> {
                DataBufferUtils.release(buffer);
                // Body not empty, back to 415..
                throw new UnsupportedMediaTypeStatusException(mediaType, this.supportedMediaTypes, elementType);
            });
            if (isBodyRequired) {
                body = body.switchIfEmpty(Mono.error(() -> handleMissingBody(bodyParam)));
            }
            return (adapter != null ? Mono.just(adapter.fromPublisher(body)) : Mono.from(body));
        }

        return Mono.error(new UnsupportedMediaTypeStatusException(mediaType, this.supportedMediaTypes, elementType));
    }

    /**
     * 核心，从对象抽取出验证的group
     *
     * @param target 目标对象
     * @return 结果
     */
    @NonNull
    private Object[] extractValidationHints(Object target) {
        ConditionalGroup group = AnnotationUtils.findAnnotation(target.getClass(), ConditionalGroup.class);
        List<Class<?>> groups = new ArrayList<>();
        if (null != group) {
            if (!group.codeless()) {
                groups.add(CodeValid.class);
            }
            if (!group.nameless()) {
                groups.add(NameValid.class);
            }
            if (ArrayUtils.isNotEmpty(group.value())) {
                for (Class<? extends ConditionalValidGroupProvider<?>> clazz : group.value()) {
                    try {
                        groups.addAll(Arrays.asList(clazz.newInstance().provide(CastUtils.cast(target))));
                    } catch (InstantiationException | IllegalAccessException ignored) {
                    }
                }
            }
        } else {
            groups = Arrays.asList(CodeValid.class, NameValid.class);
        }
        if (CollectionUtils.isNotEmpty(groups)) {
            return groups.toArray();
        }
        return new Object[]{};
    }

    private void validate(Object target, MethodParameter param,
                          BindingContext binding, ServerWebExchange exchange) {
        Object[] validationHints = extractValidationHints(target);
        String name = Conventions.getVariableNameForParameter(param);
        WebExchangeDataBinder binder = binding.createDataBinder(exchange, target, name);
        binder.validate(validationHints);
        if (binder.getBindingResult().hasErrors()) {
            throw new WebExchangeBindException(param, binder.getBindingResult());
        }
    }

    private Throwable handleReadError(MethodParameter parameter, Throwable ex) {
        return (ex instanceof DecodingException ?
                new ServerWebInputException("Failed to read HTTP message", parameter, ex) : ex);
    }

    private ServerWebInputException handleMissingBody(MethodParameter parameter) {
        String paramInfo = parameter.getExecutable().toGenericString();
        return new ServerWebInputException("Request body is missing: " + paramInfo, parameter);
    }
}
