package com.flyfish.framework.service.impl;

import com.flyfish.framework.auditor.BeanAuditor;
import com.flyfish.framework.auditor.BeanPoster;
import com.flyfish.framework.bean.SyncVo;
import com.flyfish.framework.domain.authorized.AuthorizedDomain;
import com.flyfish.framework.domain.base.AuditDomain;
import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.domain.base.Qo;
import com.flyfish.framework.repository.DefaultRepository;
import com.flyfish.framework.service.BaseService;
import com.flyfish.framework.utils.Assert;
import com.flyfish.framework.utils.CopyUtils;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.repository.core.EntityInformation;

import java.util.*;
import java.util.stream.Collectors;

/**
 * 基本的Service实现
 *
 * @param <T>
 */
public class BaseServiceImpl<T extends Domain> implements BaseService<T> {

    @Autowired
    protected DefaultRepository<T> repository;
    @Autowired(required = false)
    protected BeanAuditor<T> auditor;
    @Autowired(required = false)
    protected BeanPoster<T> poster;
    @Autowired
    private BeanAuditor<Domain> commonAuditor;
    @Autowired
    private BeanAuditor<AuditDomain> operationAuditor;
    @Autowired
    private BeanAuditor<AuthorizedDomain> authorizeAuditor;

    /**
     * 查询
     *
     * @param query 查询模型
     * @return 结果
     */
    @Override
    public Optional<T> getOne(Qo<T> query) {
        return repository.findOne(query);
    }

    /**
     * 通过Id查询
     *
     * @param id 主键
     * @return 结果
     */
    @Override
    public Optional<T> getById(String id) {
        return repository.findById(id);
    }

    /**
     * 通过实体名称访问
     *
     * @param name 名称
     * @return 结果
     */
    @Override
    public Optional<T> getByName(String name) {
        return repository.findByName(name);
    }

    /**
     * 根据ID集合来查询
     *
     * @param ids id列表
     * @return 结果
     */
    @Override
    public List<T> getByIds(Collection<String> ids) {
        return (List<T>) repository.findAllById(ids);
    }

    @Override
    public <K> List<T> getByValues(String key, List<K> values) {
        return repository.findAllByValues(key, values);
    }

    /**
     * 查询列表
     *
     * @param query 查询
     * @return 结果
     */
    @Override
    public List<T> getList(Qo<T> query) {
        return (List<T>) repository.findAll(query);
    }

    /**
     * 分页查询列表
     *
     * @param query 查询实体
     * @return 结果
     */
    @Override
    public Page<T> getPageList(Qo<T> query) {
        return repository.findAll(query, query.getPageable());
    }

    /**
     * 获取所有对象
     *
     * @return 结果
     */
    @Override
    public List<T> getAll() {
        return repository.findAll();
    }

    /**
     * 查询总记录数
     *
     * @return 总数
     */
    @Override
    public Long countAll() {
        return repository.count();
    }

    /**
     * 查询总记录数
     *
     * @param query 查询实体
     * @return 结果
     */
    @Override
    public Long count(Qo<T> query) {
        return repository.count(query);
    }

    /**
     * 插入新记录
     *
     * @param entity 实体
     */
    @Override
    public T create(T entity) {
        audit(entity);
        return post(repository.insert(entity));
    }

    /**
     * 插入 不插入null字段
     *
     * @param entity 实体
     */
    @Override
    public T createSelective(T entity) {
        audit(entity);
        return post(repository.insert(entity));
    }

    /**
     * 删除
     *
     * @param entity 实体
     */
    @Override
    public void delete(T entity) {
        repository.delete(entity);
    }

    /**
     * 根据Id删除
     *
     * @param id 主键
     */
    @Override
    public void deleteById(String id) {
        repository.deleteById(id);
    }

    /**
     * 根据id更新
     *
     * @param entity 实体
     */
    @Override
    public T updateById(T entity) {
        audit(entity);
        return post(repository.save(entity));
    }

    /**
     * 不update null
     *
     * @param entity 实体
     */
    @Override
    public T updateSelectiveById(T entity) {
        Assert.hasText(entity.getId(), "更新的主键不可为空！");
        Optional<T> saved = repository.findById(entity.getId());
        Assert.isTrue(saved.isPresent(), "要更新的信息不存在！");
        return updateById(CopyUtils.copyProps(entity, saved.get()));
    }

    /**
     * 根据ID集合批量删除
     *
     * @param ids id结合
     */
    @Override
    public void deleteBatchByIds(List<String> ids) {
        ids.forEach(this::deleteById);
    }

    /**
     * 通过某一个键删除所有
     *
     * @param qo 查询实体
     */
    @Override
    public void deleteAll(Qo<T> qo) {
        Iterable<T> list = repository.findAll(qo);
        repository.deleteAll(list);
    }

    /**
     * 删除全部实体，危险！
     */
    @Override
    public void deleteAll() {
        repository.deleteAll();
    }

    /**
     * 批量更新
     *
     * @param entities 要批量更新的集合
     */
    @Override
    public List<T> updateBatch(List<T> entities) {
        if (CollectionUtils.isNotEmpty(entities)) {
            //        entities.forEach(item -> Assert.hasText(item.getId(), "每个对象的id必须附带！"));
            // 带有id的集合
            List<String> ids = entities.stream()
                    .filter(Objects::nonNull)
                    .map(Domain::getId)
                    .filter(Objects::nonNull)
                    .collect(Collectors.toList());
            // 过滤不存在的记录
            Map<String, T> savedMap = CollectionUtils.isNotEmpty(ids) ? ((List<T>) repository.findAllById(ids)).stream()
                    .collect(Collectors.toMap(Domain::getId, t -> t)) : Collections.emptyMap();
            List<T> updating = entities.stream()
                    .map(t -> {
                        // 补全已保存信息
                        if (savedMap.containsKey(t.getId())) {
                            return CopyUtils.copyProps(t, savedMap.get(t.getId()));
                        }
                        audit(t);
                        return t;
                    })
                    .collect(Collectors.toList());
            return repository.saveAll(updating);
        }
        return entities;
    }

    /**
     * 同步数据
     *
     * @param entities 数据们
     * @return 结果
     */
    @Override
    public SyncVo<T> sync(List<T> entities) {
        List<T> results = this.updateBatch(entities);
        return SyncVo.<T>builder()
                .success(results.size())
                .list(results)
                .build();
    }

    /**
     * 对象审查
     *
     * @param entity 实体
     */
    protected void audit(T entity) {
        // 自定义审查
        if (auditor != null) {
            auditor.audit(entity);
        }
        // 公共审查
        commonAuditor.audit(entity);
        // 用户审查
        if (entity instanceof AuditDomain) {
            operationAuditor.audit((AuditDomain) entity);
        }
        // 权限审查
        if (entity instanceof AuthorizedDomain) {
            authorizeAuditor.audit((AuthorizedDomain) entity);
        }
    }

    /**
     * 后置审查
     *
     * @param entity 实体
     * @return 结果
     */
    protected T post(T entity) {
        if (null != poster) {
            poster.post(entity);
        }
        return entity;
    }

    @SuppressWarnings("unchecked")
    public <R extends DefaultRepository<T>> R getRepository() {
        return (R) repository;
    }

    /**
     * 获取集合名称
     *
     * @return 结果
     */
    @Override
    public EntityInformation<? extends Domain, String> getEntityInformation() {
        return repository.getEntityInformation();
    }
}
